@@ -1,4 +1,3 @@
-(* $Id: ag_main.ml 52381 2010-11-24 22:40:21Z martin $ *)
 
 open Printf
 
@@ -25,7 +24,7 @@ let set_once varname var x =
 Command-line parameter %S is set multiple times
 to incompatible values."
                       varname)
-        
+
     | None ->
         var := Some x
 
@@ -54,6 +53,7 @@ let main () =
   let out_prefix = ref None in
   let mode = ref (None : mode option) in
   let std_json = ref false in
+  let j_preprocess_input = ref None in
   let j_defaults = ref false in
   let unknown_field_handler = ref None in
   let type_aliases = ref None in
@@ -141,6 +141,14 @@ let main () =
           [deprecated in favor of -j-std]
           Same as -j-std.";
 
+    "-j-pp",
+    Arg.String (fun s -> set_once "-j-pp" j_preprocess_input s),
+    "<func>
+          OCaml function of type (string -> string) applied on the input
+          of each *_of_string function generated by atdgen (JSON mode).
+          This is originally intended for UTF-8 validation of the input
+          which is not performed by atdgen.";
+
     "-j-defaults",
     Arg.Set j_defaults,
     "
@@ -164,7 +172,7 @@ let main () =
         set_once "unknown field handler" unknown_field_handler s
     ),
     "FUNCTION
-          Call the given function of type (string -> unit) 
+          Call the given function of type (string -> unit)
           for every unknown JSON field found in the input
           instead of simply skipping them.
           See also -j-strict-fields.";
@@ -195,7 +203,7 @@ let main () =
                         set_once "no function definitions" with_fundefs false),
     "
           Do not dump OCaml function definitions";
-    
+
     "-ntd", Arg.Unit (fun () ->
                         set_once "no type definitions" with_typedefs false),
     "
@@ -236,12 +244,12 @@ Recommended usage: %s (-t|-b|-j|-v|-dep|-list) example.atd" Sys.argv.(0) in
   if (!std_json || !unknown_field_handler <> None) && !mode = None then
     set_once "output mode" mode `Json;
 
-  let mode = 
+  let mode =
     match !mode with
         None -> `Biniou
       | Some x -> x
   in
-  
+
   let with_create =
     match !with_create with
         Some x -> x
@@ -256,9 +264,9 @@ Recommended usage: %s (-t|-b|-j|-v|-dep|-list) example.atd" Sys.argv.(0) in
   let force_defaults =
     match mode with
         `J | `Json -> !j_defaults
-      | `T 
+      | `T
       | `B | `Biniou
-      | `V | `Validate 
+      | `V | `Validate
       | `Dep | `List -> false (* don't care *)
   in
 
@@ -295,7 +303,7 @@ Recommended usage: %s (-t|-b|-j|-v|-dep|-list) example.atd" Sys.argv.(0) in
                | _ -> base
             )
   in
-  let type_aliases = 
+  let type_aliases =
     match base_prefix with
         None ->
           (match mode with
@@ -307,7 +315,8 @@ Recommended usage: %s (-t|-b|-j|-v|-dep|-list) example.atd" Sys.argv.(0) in
               Some _ as x -> x
             | None ->
                 (match mode with
-                     `B | `J | `V -> Some (String.capitalize base ^ "_t")
+                     `B | `J | `V ->
+                       Some (String.capitalize (Filename.basename base) ^ "_t")
                    | _ -> None
           )
   in
@@ -320,7 +329,7 @@ Recommended usage: %s (-t|-b|-j|-v|-dep|-list) example.atd" Sys.argv.(0) in
       `Dep -> print_deps (get_base_prefix ())
     | `List -> print_file_list (get_base_prefix ())
     | `T | `B | `J | `V | `Biniou | `Json | `Validate ->
-        
+
         let opens = List.rev !opens in
         let make_ocaml_files =
           match mode with
@@ -332,12 +341,13 @@ Recommended usage: %s (-t|-b|-j|-v|-dep|-list) example.atd" Sys.argv.(0) in
                 Ag_oj_emit.make_ocaml_files
                   ~std: !std_json
                   ~unknown_field_handler: !unknown_field_handler
+                  ~preprocess_input: !j_preprocess_input
             | `V | `Validate ->
                 Ag_ov_emit.make_ocaml_files
             | _ -> assert false
         in
         let with_default default = function None -> default | Some x -> x in
-        
+
         make_ocaml_files
           ~opens
           ~with_typedefs: (with_default true !with_typedefs)
@@ -349,7 +359,7 @@ Recommended usage: %s (-t|-b|-j|-v|-dep|-list) example.atd" Sys.argv.(0) in
           ~type_aliases
           ~force_defaults
           atd_file ocaml_prefix
-    
+
 let () =
   try main ()
   with
