@@ -1,5 +1,3 @@
-(* $Id: ag_ob_emit.ml 52975 2010-12-21 00:48:38Z martin $ *)
-
 (*
   OCaml code generator for the biniou format.
 *)
@@ -29,7 +27,7 @@ let make_ocaml_biniou_intf ~with_create buf deref defs =
         let writer_params =
           String.concat "" (
             List.map
-              (fun s -> 
+              (fun s ->
                  sprintf "\n  Bi_io.node_tag ->\
                           \n  (Bi_outbuf.t -> '%s -> unit) ->\
                           \n  (Bi_outbuf.t -> '%s -> unit) ->" s s)
@@ -40,7 +38,7 @@ let make_ocaml_biniou_intf ~with_create buf deref defs =
           String.concat "" (
             List.map (
               fun s ->
-                sprintf 
+                sprintf
                   "\n  (Bi_io.node_tag -> (Bi_inbuf.t -> '%s)) ->\
                    \n  (Bi_inbuf.t -> '%s) ->" s s
             )
@@ -81,7 +79,7 @@ val write_%s :%s
 	bprintf buf "\
 val string_of_%s :%s
   ?len:int -> %s -> string
-  (** Serialize a value of type {!%s} into 
+  (** Serialize a value of type {!%s} into
       a biniou string. *)
 
 "
@@ -116,7 +114,7 @@ val read_%s :%s
 val %s_of_string :%s
   ?pos:int -> string -> %s
   (** Deserialize a biniou value of type {!%s}.
-      @param pos specifies the position where 
+      @param pos specifies the position where
                  reading starts. Default: 0. *)
 
 "
@@ -134,11 +132,11 @@ val %s_of_string :%s
   ) (flatten defs)
 
 
-let get_biniou_tag (x : ob_mapping) =
+let rec get_biniou_tag (x : ob_mapping) =
   match x with
       `Unit (loc, `Unit, `Unit) -> "Bi_io.unit_tag"
     | `Bool (loc, `Bool, `Bool) -> "Bi_io.bool_tag"
-    | `Int (loc, `Int o, `Int b) -> 
+    | `Int (loc, `Int o, `Int b) ->
 	(match b with
 	     `Uvint -> "Bi_io.uvint_tag"
 	   | `Svint -> "Bi_io.svint_tag"
@@ -147,7 +145,11 @@ let get_biniou_tag (x : ob_mapping) =
 	   | `Int32 -> "Bi_io.int32_tag"
 	   | `Int64 -> "Bi_io.int64_tag"
 	)
-    | `Float (loc, `Float, `Float) -> "Bi_io.float64_tag"
+    | `Float (loc, `Float, `Float b) ->
+        (match b with
+            `Float32 -> "Bi_io.float32_tag"
+          | `Float64 -> "Bi_io.float64_tag"
+        )
     | `String (loc, `String, `String) -> "Bi_io.string_tag"
     | `Sum (loc, a, `Sum x, `Sum) -> "Bi_io.variant_tag"
     | `Record (loc, a, `Record o, `Record) -> "Bi_io.record_tag"
@@ -157,8 +159,11 @@ let get_biniou_tag (x : ob_mapping) =
 	     `Array -> "Bi_io.array_tag"
 	   | `Table -> "Bi_io.table_tag"
 	)
-    | `Option (loc, x, `Option, `Option) -> "Bi_io.num_variant_tag"
+    | `Option (loc, x, `Option, `Option)
+    | `Nullable (loc, x, `Nullable, `Nullable) -> "Bi_io.num_variant_tag"
     | `Shared (loc, id, x, `Shared _, `Shared) -> "Bi_io.shared_tag"
+    | `Wrap (loc, x, `Wrap _, `Wrap) -> get_biniou_tag x
+
     | `Name (loc, s, args, None, None) -> sprintf "%s_tag" s
     | `External (loc, s, args,
                  `External (types_module, main_module, ext_name),
@@ -184,7 +189,7 @@ let get_fields deref a =
 		    `With_default ->
 		      (match o.Ag_ocaml.ocaml_default with
 			   None ->
-			     let d = 
+			     let d =
 			       Ag_ocaml.get_implicit_ocaml_default
                                  deref x.f_value in
 			     if d = None then
@@ -249,9 +254,11 @@ let rec get_writer_name
 	       error loc "Unsupported combination of OCaml/Biniou int types"
 	)
 
-    | `Float (loc, `Float, `Float) ->
-	sprintf "Bi_io.write_%sfloat64" un
-
+    | `Float (loc, `Float, `Float b) ->
+        (match b with
+            `Float32 -> sprintf "Bi_io.write_%sfloat32" un
+          | `Float64 -> sprintf "Bi_io.write_%sfloat64" un
+        )
     | `String (loc, `String, `String) ->
 	sprintf "Bi_io.write_%sstring" un
 
@@ -351,7 +358,11 @@ let rec get_reader_name
 	       error loc "Unsupported combination of OCaml/Biniou int types"
 	)
 
-    | `Float (loc, `Float, `Float) -> xreader "float"
+    | `Float (loc, `Float, `Float b) ->
+        (match b with
+            `Float32 -> xreader "float32"
+          | `Float64 -> xreader "float64"
+        )
 
     | `String (loc, `String, `String) -> xreader "string"
 
@@ -475,7 +486,7 @@ let rec make_writer ~tagged deref (x : ob_mapping) : Ag_indent.t list =
 		]
 	    ) a
 	  in
-	  [ 
+	  [
             `Line (sprintf "Bi_vint.write_uvint ob %i;" len);
 	    `Inline (List.flatten (Array.to_list a))
           ]
@@ -523,7 +534,8 @@ let rec make_writer ~tagged deref (x : ob_mapping) : Ag_indent.t list =
                ]
 	)
 
-    | `Option (loc, x, `Option, `Option) ->
+    | `Option (loc, x, `Option, `Option)
+    | `Nullable (loc, x, `Nullable, `Nullable) ->
 	[
           `Line (sprintf "Ag_ob_run.write_%soption (" un);
 	  `Block (make_writer ~tagged:true deref x);
@@ -542,6 +554,21 @@ let rec make_writer ~tagged deref (x : ob_mapping) : Ag_indent.t list =
           `Line ")";
         ]
 
+    | `Wrap (loc, x, `Wrap o, `Wrap) ->
+        let simple_writer = make_writer ~tagged deref x in
+        (match o with
+            None -> simple_writer
+          | Some { Ag_ocaml.ocaml_wrap_t; ocaml_wrap; ocaml_unwrap } ->
+              [
+                `Line "fun ob x -> (";
+                `Block [
+                  `Line (sprintf "let x = ( %s ) x in (" ocaml_unwrap);
+                  `Block simple_writer;
+                  `Line ") ob x)";
+                ]
+              ]
+        )
+
     | _ -> assert false
 
 
@@ -583,13 +610,13 @@ and make_record_writer deref tagged a record_kind =
     (* count the number of defined optional fields in order
        to determine the length of the record *)
     let min_len =
-      List.fold_left 
+      List.fold_left
 	(fun n (_, _, _, opt, _) -> if opt then n else n + 1) 0 fields
     in
     let max_len = List.length fields in
     if min_len = max_len then
       [ `Line (sprintf "Bi_vint.write_uvint ob %i;" max_len) ]
-    else 
+    else
       [
         (* Using a ref because many "let len = ... len + 1 in"
            cause ocamlopt to take a very long time to finish *)
@@ -611,7 +638,7 @@ and make_record_writer deref tagged a record_kind =
 	`Line "Bi_vint.write_uvint ob !len;"
       ]
   in
-  
+
   let write_fields =
     List.map (
       fun (x, ocaml_fname, ocaml_default, optional, unwrapped) ->
@@ -668,7 +695,7 @@ and make_table_writer deref tagged list_kind x =
   let a, record_kind =
     match deref x with
 	`Record (_, a, `Record record_kind, `Record) -> a, record_kind
-      | _ -> 
+      | _ ->
 	  error (loc_of_mapping x) "Not a record type"
   in
   let dot =
@@ -734,7 +761,7 @@ let study_record deref fields =
   let maybe_constant =
     List.for_all (function (_, _, Some _, _, _) -> true | _ -> false) fields
   in
-  let _, init_fields = 
+  let _, init_fields =
     List.fold_right (
       fun (x, name, default, opt, unwrap) (maybe_constant, l) ->
 	let maybe_constant, v =
@@ -742,7 +769,7 @@ let study_record deref fields =
 	      None ->
 		assert (not opt);
 		(*
-		  The initial value is a float because the record may be 
+		  The initial value is a float because the record may be
 		  represented as a double_array (unboxed floats).
 		  Float values work in all cases.
 		*)
@@ -769,7 +796,7 @@ let study_record deref fields =
   let init_val = [ `Line "{"; `Block init_fields; `Line "}" ] in
 
   let k = n / 31 + (if n mod 31 > 0 then 1 else 0) in
-  let init_bits = 
+  let init_bits =
     Array.to_list (
       Array.init k (
 	fun i -> `Line (sprintf "let bits%i = ref 0 in" i)
@@ -788,7 +815,7 @@ let study_record deref fields =
   let set_bit z0 =
     match mapping.(z0) with
 	None -> []
-      | Some z ->    
+      | Some z ->
 	  let i = z / 31 in
 	  let j = z mod 31 in
 	  [ `Line (sprintf "bits%i := !bits%i lor 0x%x;" i i (1 lsl j)) ]
@@ -857,7 +884,7 @@ let wrap_bodies ~tagged l =
     let cases =
       List.map (
 	fun (expected_tag, body) ->
-	  `Inline [ 
+	  `Inline [
 	    `Line (sprintf "| %i -> " expected_tag);
 	    `Block body;
 	  ]
@@ -877,9 +904,9 @@ let wrap_bodies ~tagged l =
     let cases =
       List.map (
 	fun (expected_tag, body) ->
-	  `Inline [ 
+	  `Inline [
 	    `Line (sprintf "| %i -> " expected_tag);
-	    `Block [ 
+	    `Block [
 	      `Line "(fun ib ->";
 	      `Block body;
 	      `Line ")";
@@ -894,7 +921,7 @@ let wrap_bodies ~tagged l =
 	`Line "| _ -> Ag_ob_run.read_error ()"
       ]
     ]
-      
+
 
 let rec make_reader deref ~tagged (x : ob_mapping) : Ag_indent.t list =
   match x with
@@ -988,7 +1015,8 @@ let rec make_reader deref ~tagged (x : ob_mapping) : Ag_indent.t list =
 				     Bi_io.array_tag, body2 ]
 	)
 
-    | `Option (loc, x, `Option, `Option) ->
+    | `Option (loc, x, `Option, `Option)
+    | `Nullable (loc, x, `Nullable, `Nullable) ->
 	let body = [
 	  `Line "match Char.code (Bi_inbuf.read_char ib) with";
 	  `Block [
@@ -1037,6 +1065,30 @@ let rec make_reader deref ~tagged (x : ob_mapping) : Ag_indent.t list =
         in
         wrap_body ~tagged Bi_io.shared_tag body
 
+    | `Wrap (loc, x, `Wrap o, `Wrap) ->
+        let simple_reader = make_reader deref ~tagged x in
+        (match o with
+            None -> simple_reader
+          | Some { Ag_ocaml.ocaml_wrap } ->
+              if tagged then
+                [
+                  `Line "fun ib ->";
+                  `Block [
+                    `Line (sprintf "( %s ) ((" ocaml_wrap);
+                    `Block simple_reader;
+                    `Line ") ib)";
+                  ]
+                ]
+              else
+                [
+                  `Line "fun tag ib ->";
+                  `Block [
+                    `Line (sprintf "( %s ) ((" ocaml_wrap);
+                    `Block simple_reader;
+                    `Line ") tag ib)";
+                  ]
+                ]
+        )
     | _ -> assert false
 
 
@@ -1066,7 +1118,7 @@ and make_variant_reader deref tick x : Ag_indent.t list =
         ]
 
 and make_record_reader ?shared_id deref ~tagged a record_kind =
-  let fields = get_fields deref a in  
+  let fields = get_fields deref a in
   let init_val, init_bits, set_bit, check_bits = study_record deref fields in
 
   let build share body =
@@ -1084,7 +1136,7 @@ and make_record_reader ?shared_id deref ~tagged a record_kind =
       `Line "Ag_ob_run.identity x"
     ]
   in
-  
+
   let loop body =
     match shared_id with
         None -> build [] body
@@ -1119,7 +1171,7 @@ and make_record_reader ?shared_id deref ~tagged a record_kind =
 	    else x.f_value
 	  in
 	  let wrap l =
-	    if unwrapped then 
+	    if unwrapped then
 	      [
 		`Line "Some (";
 		`Block l;
@@ -1154,7 +1206,7 @@ and make_record_reader ?shared_id deref ~tagged a record_kind =
       ]
     ]
   in
-      
+
   loop body
 
 and make_tuple_reader deref ~tagged a =
@@ -1188,7 +1240,7 @@ and make_tuple_reader deref ~tagged a =
 	  fun i (x, default) ->
 	    let read_value = make_reader deref ~tagged:true x.cel_value in
 	    let get_value =
-	      if i < min_length then 
+	      if i < min_length then
                 [
                   `Line "(";
                   `Block read_value;
@@ -1214,7 +1266,7 @@ and make_tuple_reader deref ~tagged a =
 	) cells
       )
     )
-  in  
+  in
 
   let make_tuple =
     sprintf "(%s)"
@@ -1278,7 +1330,7 @@ and make_table_reader deref loc list_kind x =
 	      `Line "in";
 	      `Line "(fun x ib ->";
 	      `Block [
-		`Line (sprintf 
+		`Line (sprintf
 			 "Obj.set_field (Obj.repr x) %i \
                             (Obj.repr (read ib)))" i
 		      )
@@ -1530,13 +1582,13 @@ let make_ml
 
 let make_ocaml_files
     ~opens
-    ~with_typedefs 
+    ~with_typedefs
     ~with_create
     ~with_fundefs
     ~all_rec
     ~pos_fname
     ~pos_lnum
-    ~type_aliases 
+    ~type_aliases
     ~force_defaults
     atd_file out =
   let head, m0 =
@@ -1569,7 +1621,7 @@ let make_ocaml_files
     Ag_ocaml.ocaml_of_atd ~target:`Biniou ~type_aliases (head, m1) in
   let ocaml_impl_misc = make_shared_id_defs m0 in
   let defs = translate_mapping m2 in
-  let header = 
+  let header =
     let src =
       match atd_file with
           None -> "stdin"
@@ -1577,7 +1629,7 @@ let make_ocaml_files
     in
     sprintf "(* Auto-generated from %s *)\n" src
   in
-  let mli = 
+  let mli =
     make_mli ~header ~opens ~with_typedefs ~with_create ~with_fundefs
       ocaml_typedefs (Ag_mapping.make_deref defs1) defs1
   in
