@@ -16,10 +16,16 @@ open Ag_mapping
 type atd_ocaml_sum = [ `Classic | `Poly ]
 type atd_ocaml_record = [ `Record | `Object ]
 
-type atd_ocaml_int = [ `Int | `Char | `Int32 | `Int64 ]
+type atd_ocaml_int = [ `Int | `Char | `Int32 | `Int64 | `Float ]
 type atd_ocaml_list = [ `List | `Array ]
 
-type atd_ocaml_shared  = [ `Flat | `Ref ]
+type atd_ocaml_shared = [ `Flat | `Ref ]
+
+type atd_ocaml_wrap = {
+  ocaml_wrap_t : string;
+  ocaml_wrap : string;
+  ocaml_unwrap : string;
+}
 
 type atd_ocaml_field = {
   ocaml_default : string option;
@@ -50,7 +56,9 @@ type atd_ocaml_repr =
     | `Tuple
     | `List of atd_ocaml_list
     | `Option
+    | `Nullable
     | `Shared of atd_ocaml_shared
+    | `Wrap of atd_ocaml_wrap option
     | `Name of string
     | `External of (string * string * string)
         (*
@@ -74,6 +82,7 @@ let ocaml_int_of_string s : atd_ocaml_int option =
     | "char" -> Some `Char
     | "int32" -> Some `Int32
     | "int64" -> Some `Int64
+    | "float" -> Some `Float
     | _ -> None
 
 let string_of_ocaml_int (x : atd_ocaml_int) =
@@ -82,6 +91,7 @@ let string_of_ocaml_int (x : atd_ocaml_int) =
     | `Char -> "Char.t"
     | `Int32 -> "Int32.t"
     | `Int64 -> "Int64.t"
+    | `Float -> "float"
 
 let ocaml_sum_of_string s : atd_ocaml_sum option =
   match s with
@@ -155,6 +165,33 @@ let get_ocaml_list an =
 let get_ocaml_shared an =
   Atd_annot.get_field ocaml_shared_of_string `Flat ["ocaml"] "repr" an
 
+let get_ocaml_wrap loc an =
+  let module_ =
+    Atd_annot.get_field (fun s -> Some (Some s)) None ["ocaml"] "module" an in
+  let default field =
+    match module_ with
+        None -> None
+      | Some s -> Some (sprintf "%s.%s" s field)
+  in
+  let t =
+    Atd_annot.get_field (fun s -> Some (Some s))
+      (default "t") ["ocaml"] "t" an
+  in
+  let wrap =
+    Atd_annot.get_field (fun s -> Some (Some s))
+      (default "wrap") ["ocaml"] "wrap" an
+  in
+  let unwrap =
+    Atd_annot.get_field (fun s -> Some (Some s))
+      (default "unwrap") ["ocaml"] "unwrap" an
+  in
+  match t, wrap, unwrap with
+      None, None, None -> None
+    | Some t, Some wrap, Some unwrap ->
+        Some { ocaml_wrap_t = t; ocaml_wrap = wrap; ocaml_unwrap = unwrap }
+    | _ ->
+        Ag_error.error loc "Incomplete annotation. Missing t, wrap or unwrap"
+
 let get_ocaml_cons default an =
   Atd_annot.get_field (fun s -> Some s) default ["ocaml"] "name" an
 
@@ -262,11 +299,18 @@ let rec map_expr (x : type_expr) : ocaml_expr =
 	`Name (s, [map_expr x])
     | `Option (loc, x, an) ->
 	`Name ("option", [map_expr x])
+    | `Nullable (loc, x, an) ->
+	`Name ("option", [map_expr x])
     | `Shared (loc, x, a) ->
         (match get_ocaml_shared a with
              `Flat -> map_expr x
            | `Ref -> `Name ("Pervasives.ref", [map_expr x])
         )
+    | `Wrap (loc, x, a) ->
+        (match get_ocaml_wrap loc a with
+            None -> map_expr x
+          | Some { ocaml_wrap_t } -> `Name (ocaml_wrap_t, [])
+        )
     | `Name (loc, (loc2, s, l), an) ->
 	let s = get_ocaml_type_path s an in
 	`Name (s, List.map map_expr l)
@@ -332,7 +376,7 @@ let map_def
     in
     if x = None && alias = None then
       None
-    else 
+    else
       Some {
         o_def_name = (s, param);
         o_def_alias = alias;
@@ -381,6 +425,8 @@ let rec ocaml_of_expr_mapping (x : (atd_ocaml_repr, _) mapping) : ocaml_expr =
         `Name (string_of_ocaml_list kind, [ocaml_of_expr_mapping x])
     | `Option (loc, x, `Option, _) ->
         `Name ("option", [ocaml_of_expr_mapping x])
+    | `Nullable (loc, x, `Nullable, _) ->
+        `Name ("option", [ocaml_of_expr_mapping x])
     | `Name (loc, s, l, _, _) ->
         `Name (s, List.map ocaml_of_expr_mapping l)
     | `Tvar (loc, s) ->
@@ -425,13 +471,13 @@ let plist = { list with
 		space_before_closing = false }
 
 let hlist = { list with wrap_body = `No_breaks }
-let shlist = { hlist with 
+let shlist = { hlist with
 		 stick_to_label = false;
 		 space_after_opening = false;
 		 space_before_closing = false }
 let shlist0 = { shlist with space_after_separator = false }
 
-let llist = { 
+let llist = {
   list with
     separators_stick_left = false;
     space_before_separator = true;
@@ -445,7 +491,7 @@ let lplist = {
 }
 
 let vseq = {
-  list with 
+  list with
     indent_body = 0;
     wrap_body = `Force_breaks;
 }
@@ -469,11 +515,11 @@ let horizontal_sequence0 l = List (("", "", "", shlist0), l)
 let rec insert sep = function
     [] | [_] as l -> l
   | x :: l -> x :: sep @ insert sep l
-      
+
 let rec insert2 f = function
     [] | [_] as l -> l
   | x :: (y :: _ as l) -> x :: f x y @ insert2 f l
-      
+
 
 let vertical_sequence ?(skip_lines = 0) l =
   let l =
@@ -538,8 +584,8 @@ let make_ocamldoc_blocks (l : Ag_doc.block list) =
     ) (l :> [ Ag_doc.block | `Before_paragraph ] list)
   in
   List.map make_ocamldoc_block l
-  
-  
+
+
 let make_ocamldoc_comment (`Text l) =
   let blocks = make_ocamldoc_blocks l in
   let xlist =
@@ -548,14 +594,14 @@ let make_ocamldoc_comment (`Text l) =
       | _ -> vlist
   in
   List (("(**", "", "*)", xlist), blocks)
-       
+
 let prepend_ocamldoc_comment doc x =
   match doc with
       None -> x
     | Some y ->
         let comment = make_ocamldoc_comment y in
         List (("", "", "", rlist), [comment;x])
-          
+
 let append_ocamldoc_comment x doc =
   match doc with
       None -> x
@@ -574,7 +620,7 @@ let rec format_module_item
     if b then s1 ^ s2
     else s1
   in
-  let part1 = 
+  let part1 =
     horizontal_sequence (
       make_atom type_ ::
         prepend_type_param param
@@ -599,7 +645,7 @@ let rec format_module_item
   let part123 =
     match expr with
         None -> part12
-            
+
       | Some t ->
 	  Label (
 	    (part12, label),
@@ -607,14 +653,14 @@ let rec format_module_item
 	  )
   in
   prepend_ocamldoc_comment doc part123
-    
-	  
+
+	
 and prepend_type_param l tl =
   match l with
       [] -> tl
     | _ ->
 	let make_var s = make_atom ("'" ^ s) in
-	let x = 
+	let x =
 	  match l with
 	      [s] -> make_var s
 	    | l -> List (("(", ",", ")", plist), List.map make_var l)
@@ -625,7 +671,7 @@ and prepend_type_args l tl =
   match l with
       [] -> tl
     | _ ->
-	let x = 
+	let x =
 	  match l with
 	      [t] -> format_type_expr t
 	    | l -> List (("(", ",", ")", plist), List.map format_type_expr l)
@@ -667,7 +713,7 @@ and format_type_expr x =
 
 and format_type_name name args =
   horizontal_sequence (prepend_type_args args [ make_atom name ])
-	  
+	
 and format_field ((s, is_mutable), t, doc) =
   let l =
     let l = [make_atom (s ^ ":")] in
@@ -707,7 +753,7 @@ and format_variant kind (s, o, doc) =
 let format_module_items is_rec (l : ocaml_module_body) =
   match l with
       x :: l ->
-	format_module_item true x :: 
+	format_module_item true x ::
 	  List.map (fun x -> format_module_item false x) l
     | [] -> []
 
@@ -752,11 +798,12 @@ let get_full_type_name x =
     | l ->
         let l = List.map (fun s -> "'" ^ s) l in
         sprintf "(%s) %s" (String.concat ", " l) s
-        
+
 
 let unwrap_option deref x =
   match deref x with
-      `Option (_, x, _, _) -> x
+      `Option (_, x, _, _)
+    | `Nullable (_, x, _, _) -> x
     | `Name (loc, s, _, _, _) ->
 	Ag_error.error loc ("Not an option type: " ^ s)
     | x ->
@@ -773,12 +820,14 @@ let get_implicit_ocaml_default deref x =
 		  `Int -> "0"
 		| `Char -> "'\000'"
 		| `Int32 -> "0l"
-		| `Int64 -> "0L")
+		| `Int64 -> "0L"
+                | `Float -> "0.")
     | `Float (loc, `Float, _) -> Some "0.0"
     | `String (loc, `String, _) -> Some "\"\""
     | `List (loc, x, `List `List, _) -> Some "[]"
     | `List (loc, x, `List `Array, _) -> Some "[||]"
     | `Option (loc, x, `Option, _) -> Some "None"
+    | `Nullable (loc, x, `Nullable, _) -> Some "None"
     | _ -> None
 
 
@@ -804,7 +853,7 @@ let map_record_creator_field deref x =
         let impl1 = sprintf "\n  ?%s" fname in
         intf, impl1, impl2
 
-    | `With_default -> 
+    | `With_default ->
         let t = ocaml_of_expr (ocaml_of_expr_mapping x.f_value) in
         let intf = sprintf "\n  ?%s: %s ->" fname t in
         let impl1 =
