@@ -25,6 +25,9 @@ type param = {
        and does something with it such as displaying a warning message. *)
 
   force_defaults : bool;
+
+  preprocess_input : string option;
+    (* intended for UTF-8 validation *)
 }
 
 
@@ -65,7 +68,7 @@ val string_of_%s :%s
   ?len:int -> %s -> string
   (** Serialize a value of type {!%s}
       into a JSON string.
-      @param len specifies the initial length 
+      @param len specifies the initial length
                  of the buffer used internally.
                  Default: 1024. *)
 
@@ -132,7 +135,7 @@ let get_fields p a =
                     `With_default ->
                       (match o.Ag_ocaml.ocaml_default with
                            None ->
-                             let d = 
+                             let d =
                                Ag_ocaml.get_implicit_ocaml_default
                                  p.deref x.f_value in
                              if d = None then
@@ -196,11 +199,22 @@ let rec get_writer_name
 	   | `Char -> "Ag_oj_run.write_int8"
 	   | `Int32 -> "Ag_oj_run.write_int32"
 	   | `Int64 -> "Ag_oj_run.write_int64"
+           | `Float -> "Ag_oj_run.write_float_as_int"
 	)
 
-    | `Float (loc, `Float, `Float) ->
-        if p.std then "Yojson.Safe.write_std_float"
-        else "Yojson.Safe.write_float"
+    | `Float (loc, `Float, `Float j) ->
+        (match j with
+            `Float None ->
+              if p.std then "Yojson.Safe.write_std_float"
+              else "Yojson.Safe.write_float"
+          | `Float (Some precision) ->
+              if p.std then
+                sprintf "Yojson.Safe.write_std_float_prec %i" precision
+              else
+                sprintf "Yojson.Safe.write_float_prec %i" precision
+          | `Int ->
+              "Ag_oj_run.write_float_as_int"
+        )
 
     | `String (loc, `String, `String) ->
 	"Yojson.Safe.write_string"
@@ -249,9 +263,10 @@ let rec get_reader_name
 	   | `Char -> "Ag_oj_run.read_int8"
 	   | `Int32 -> "Ag_oj_run.read_int32"
 	   | `Int64 -> "Ag_oj_run.read_int64"
+           | `Float -> "Ag_oj_run.read_number"
 	)
 
-    | `Float (loc, `Float, `Float) -> "Ag_oj_run.read_number"
+    | `Float (loc, `Float, `Float j) -> "Ag_oj_run.read_number"
 
     | `String (loc, `String, `String) -> "Ag_oj_run.read_string"
 
@@ -339,14 +354,14 @@ let rec make_writer p (x : oj_mapping) : Ag_indent.t list =
 	      ]
 	  ) a
 	in
-	let l = 
+	let l =
 	  insert (`Line "Bi_outbuf.add_char ob ',';") (Array.to_list a)
 	in
         let op, cl =
           if p.std then '[', ']'
           else '(', ')'
         in
-	[ 
+	[
 	  `Annot ("fun", `Line "fun ob x ->");
 	  `Block [
 	    `Line (sprintf "Bi_outbuf.add_char ob %C;" op);
@@ -384,21 +399,42 @@ let rec make_writer p (x : oj_mapping) : Ag_indent.t list =
 	)
 
     | `Option (loc, x, `Option, `Option) ->
-	[ 
+	[
 	  `Line (sprintf "Ag_oj_run.write_%soption ("
                    (if p.std then "std_" else ""));
 	  `Block (make_writer p x);
 	  `Line ")";
 	]
 
+    | `Nullable (loc, x, `Nullable, `Nullable) ->
+	[
+	  `Line "Ag_oj_run.write_nullable (";
+	  `Block (make_writer p x);
+	  `Line ")";
+	]
+
     | `Shared (loc, _, _, _, _) ->
         error loc "Sharing is not supported by the JSON interface"
 
+    | `Wrap (loc, x, `Wrap o, `Wrap) ->
+        (match o with
+            None -> make_writer p x
+          | Some { Ag_ocaml.ocaml_wrap_t; ocaml_wrap; ocaml_unwrap } ->
+              [
+                `Line "fun ob x -> (";
+                `Block [
+                  `Line (sprintf "let x = ( %s ) x in (" ocaml_unwrap);
+                  `Block (make_writer p x);
+                  `Line ") ob x)";
+                ]
+              ]
+        )
+
     | _ -> assert false
 
 
 
-and make_variant_writer p tick x : 
+and make_variant_writer p tick x :
     Ag_indent.t list =
   let o, j =
     match x.var_arepr, x.var_brepr with
@@ -504,7 +540,7 @@ and make_record_writer p a record_kind =
     `Line "Bi_outbuf.add_char ob '{';";
     `Line "let is_first = ref true in";
     `Inline write_fields;
-    `Line "Bi_outbuf.add_char ob '}';";    
+    `Line "Bi_outbuf.add_char ob '}';";
   ]
 
 
@@ -513,7 +549,7 @@ let study_record deref fields =
   let maybe_constant =
     List.for_all (function (_, _, Some _, _, _, _) -> true | _ -> false) fields
   in
-  let _, init_fields = 
+  let _, init_fields =
     List.fold_right (
       fun (x, oname, default, jname, opt, unwrap) (maybe_constant, l) ->
 	let maybe_constant, v =
@@ -521,7 +557,7 @@ let study_record deref fields =
 	      None ->
 		assert (not opt);
 		(*
-                  The initial value is a float because the record may be 
+                  The initial value is a float because the record may be
                   represented as a double_array (unboxed floats).
                   Float values work in all cases.
 		*)
@@ -548,7 +584,7 @@ let study_record deref fields =
   let init_val = [ `Line "{"; `Block init_fields; `Line "}" ] in
 
   let k = n / 31 + (if n mod 31 > 0 then 1 else 0) in
-  let init_bits = 
+  let init_bits =
     Array.to_list (
       Array.init k (
 	fun i -> `Line (sprintf "let bits%i = ref 0 in" i)
@@ -567,7 +603,7 @@ let study_record deref fields =
   let set_bit z0 =
     match mapping.(z0) with
 	None -> []
-      | Some z ->    
+      | Some z ->
 	  let i = z / 31 in
 	  let j = z mod 31 in
 	  [ `Line (sprintf "bits%i := !bits%i lor 0x%x;" i i (1 lsl j)) ]
@@ -772,9 +808,36 @@ let rec make_reader p (x : oj_mapping) : Ag_indent.t list =
 	in
 	make_reader p (`Sum (loc, a, `Sum `Classic, `Sum))
 
+    | `Nullable (loc, x, `Nullable, `Nullable) ->
+        [
+          `Line "fun p lb ->";
+          `Block [
+            `Line "Yojson.Safe.read_space p lb;";
+            `Line "if Yojson.Safe.read_null_if_possible p lb then None";
+            `Line "else Some ((";
+            `Block (make_reader p x);
+            `Line ") p lb)"
+          ]
+        ]
+
     | `Shared (loc, _, _, _, _) ->
         error loc "Sharing is not supported by the JSON interface"
 
+    | `Wrap (loc, x, `Wrap o, `Wrap) ->
+        (match o with
+            None -> make_reader p x
+          | Some { Ag_ocaml.ocaml_wrap_t; ocaml_wrap; ocaml_unwrap } ->
+              [
+                `Line "fun p lb ->";
+                `Block [
+                  `Line "let x = (";
+                  `Block (make_reader p x);
+                  `Line ") p lb in";
+                  `Line (sprintf "( %s ) x" ocaml_wrap);
+                ]
+              ]
+        )
+
     | _ -> assert false
 
 
@@ -833,7 +896,7 @@ and make_variant_reader p tick std x : (string * Ag_indent.t list) =
 
 
 and make_record_reader p loc a record_kind =
-  let fields = get_fields p a in  
+  let fields = get_fields p a in
   let init_val, init_bits, set_bit, check_bits = study_record p.deref fields in
 
   let read_field =
@@ -846,7 +909,7 @@ and make_record_reader p loc a record_kind =
 	    else x.f_value
 	  in
 	  let wrap l =
-            if unwrapped then 
+            if unwrapped then
               [
 		`Line "Some (";
 		`Block l;
@@ -1038,7 +1101,7 @@ and make_tuple_reader deref a =
 	) cells
       )
     )
-  in  
+  in
 
   let make_tuple =
     sprintf "(%s)"
@@ -1054,7 +1117,7 @@ and make_tuple_reader deref a =
     done;
     sprintf "[ %s ]" (String.concat "; " !acc)
   in
-  
+
   let finish_empty_tuple =
     if min_length = 0 then
       [
@@ -1149,12 +1212,18 @@ let make_ocaml_json_reader p is_rec let1 let2 def =
     if is_function reader_expr || not is_rec then "", ""
     else " p lb", " p lb"
   in
+  let pp =
+    match p.preprocess_input with
+        None -> []
+      | Some f -> [ `Line (sprintf "let s = ( %s ) s in" f) ]
+  in
   [
     `Line (sprintf "%s %s%s = (" let1 read extra_param);
     `Block (List.map Ag_indent.strip reader_expr);
     `Line (sprintf ")%s" extra_args);
     `Line (sprintf "%s %s s =" let2 of_string);
     `Block [
+      `Inline pp;
       `Line (
         sprintf "%s (Yojson.Safe.init_lexer ()) \
                        (Lexing.from_string s)" read);
@@ -1176,12 +1245,14 @@ let get_let ~is_rec ~is_first =
 
 let make_ocaml_json_impl
     ~std ~unknown_field_handler ~with_create ~force_defaults
+    ~preprocess_input
     buf deref defs =
   let p = {
     deref = deref;
     std = std;
     unknown_field_handler = unknown_field_handler;
     force_defaults = force_defaults;
+    preprocess_input;
   } in
   let ll =
     List.map (
@@ -1246,7 +1317,7 @@ let make_mli
 
 let make_ml
     ~header ~opens ~with_typedefs ~with_create ~with_fundefs
-    ~std ~unknown_field_handler ~force_defaults
+    ~std ~unknown_field_handler ~force_defaults ~preprocess_input
     ocaml_typedefs deref defs =
   let buf = Buffer.create 1000 in
   bprintf buf "%s\n" header;
@@ -1257,7 +1328,8 @@ let make_ml
     bprintf buf "\n";
   if with_fundefs then
     make_ocaml_json_impl
-      ~std ~unknown_field_handler ~with_create ~force_defaults buf deref defs;
+      ~std ~unknown_field_handler ~with_create ~force_defaults
+      ~preprocess_input buf deref defs;
   Buffer.contents buf
 
 let make_ocaml_files
@@ -1272,6 +1344,7 @@ let make_ocaml_files
     ~pos_lnum
     ~type_aliases
     ~force_defaults
+    ~preprocess_input
     atd_file out =
   let head, m0 =
     match atd_file with
@@ -1301,7 +1374,7 @@ let make_ocaml_files
   let ocaml_typedefs =
     Ag_ocaml.ocaml_of_atd ~target:`Json ~type_aliases (head, m1) in
   let defs = translate_mapping m2 in
-  let header = 
+  let header =
     let src =
       match atd_file with
           None -> "stdin"
@@ -1309,13 +1382,13 @@ let make_ocaml_files
     in
     sprintf "(* Auto-generated from %s *)\n" src
   in
-  let mli = 
+  let mli =
     make_mli ~header ~opens ~with_typedefs ~with_create ~with_fundefs
       ocaml_typedefs (Ag_mapping.make_deref defs1) defs1
   in
   let ml =
     make_ml ~header ~opens ~with_typedefs ~with_create ~with_fundefs
-      ~std ~unknown_field_handler ~force_defaults
+      ~std ~unknown_field_handler ~force_defaults ~preprocess_input
       ocaml_typedefs (Ag_mapping.make_deref defs) defs
   in
   Ag_ox_emit.write_ocaml out mli ml
